
proc command:insert {w arg text} {
	global	cutbuffer

	switch -exact $arg {
		text		{ puttext $w $text }
		newline		{ insertnewline $w }
		tab		{ inserttab $w }
		brace		{ insblinkbrace $w $text }
		cutbuffer	{ command [list insert text $cutbuffer] }
		coms		{ play_comlist $text }
		filter		{ filter_text $text }
		privat		{ insert_privat $text }
		exec		{ exec_prog $text }
		undo		{ doundo }
		redo		{ doredo }
		bookmark	{ bookmark_set $w $text insert }
		char		{ insert_special_char $w $text }
		default		{ showinfo "Invalid command <insert $arg>." }
	}
}

proc command:help {w arg addr} {
	global	ownpath

	switch -exact $arg {
		document	{ command showhelp }
		copyright	{ file_goto_named $ownpath/doc/COPYING }
		default		{ showinfo "Invalid command <help $arg>." }
	}
}

proc command:showhelp {w arg addr} {
	global ownpath programs
	eval exec "$programs(html) file://${ownpath}/doc/eddi.html &"
}

proc command:move {w arg text} {
	switch -exact $arg {
		left		{ if {$text == {}} {tkTextSetCursor $w {insert-1c}} else {tkTextSetCursor $w "insert-$text c"} }
		right		{ if {$text == {}} {tkTextSetCursor $w {insert+1c}} else {tkTextSetCursor $w "insert+$text c"} }
		up		{ if {$text == {}} {tkTextSetCursor $w {insert-1l}} else {tkTextSetCursor $w "insert-$text l"} }
		down		{ if {$text == {}} {tkTextSetCursor $w {insert+1l}} else {tkTextSetCursor $w "insert+$text l"} }
		linestart	{ tkTextSetCursor $w {insert linestart} }
		lineend		{ tkTextSetCursor $w {insert lineend} }
		filestart	{ tkTextSetCursor $w {1.0} }
		fileend		{ tkTextSetCursor $w {end} }
		pageup		{ set ix [tkTextScrollPages $w -1]; if {[$w compare $ix == insert]} {set ix 1.0}; tkTextSetCursor $w $ix }
		pagedown	{ set ix [tkTextScrollPages $w 1]; if {[$w compare $ix == insert]} {set ix end}; tkTextSetCursor $w $ix}
		prevword	{ tkTextSetCursor $w [indexprevword $w insert] }
		nextword	{ tkTextSetCursor $w [indexnextword $w insert] }
		prevpar		{ tkTextSetCursor $w [tkTextPrevPara $w insert] }
		nextpar		{ tkTextSetCursor $w [tkTextNextPara $w insert] }
		default		{ showinfo "Invalid command <move $arg>." }
	}
}

proc command:goto {w arg num} {
	switch -exact $arg {
		panel		{ gotoline {} }
		file		{ file_goto_named $num }
		fileno		{ project_goto_file $num }
		bookmark	{ bookmark_goto $w $num }
		line		{ gotoline $num }
		default		{ showinfo "Invalid command <goto $arg>." }
	}
}

proc command:mark {w arg} {
	switch -exact $arg {
		left		{ tkTextKeySelect $w {insert-1c} }
		right		{ tkTextKeySelect $w {insert+1c} }
		up		{ tkTextKeySelect $w {insert-1l} }
		down		{ tkTextKeySelect $w {insert+1l} }
		linestart	{ tkTextKeySelect $w {insert linestart} }
		lineend		{ tkTextKeySelect $w {insert lineend} }
		filestart	{ tkTextKeySelect $w {1.0} }
		fileend		{ tkTextKeySelect $w {end} }
		pageup		{ tkTextKeySelect $w [tkTextScrollPages $w 1] }
		pagedown	{ tkTextKeySelect $w [tkTextScrollPages $w -1] }
		prevword	{ tkTextKeySelect $w [indexprevword $w insert] }
		nextword	{ tkTextKeySelect $w [indexnextword $w insert] }
		prevpar		{ tkTextKeySelect $w [tkTextPrevPara $w insert] }
		nextpar		{ tkTextKeySelect $w [tkTextNextPara $w insert] }
		setstart	{ $w mark set anchor insert }
		setend		{ set tkPriv(selectMode) char; tkTextKeyExtend $w insert }
		default		{ showinfo "Invalid command <mark $arg>." }
	}
}

proc command:wssel {w arg} {
	switch -exact $arg {
		begin		{ wssel_begin $w insert }
		end		{ wssel_end $w insert }
		copy		{ wssel_copy $w }
		move		{ wssel_move $w }
		del		{ wssel_del }
		toggle		{ wssel_toggle }
		default		{ showinfo "Invalid command <mark $arg>." }
	}
}

proc command:delete {w arg} {
	switch -exact $arg {
		char		{ delregion $w insert {insert+1c} }
		back		{ if {[$w index insert] != {1.0}} {delregion $w [$w index {insert-1c}] insert} }
		line		{ delregion $w [$w index {insert linestart}] [$w index {insert+1l linestart}] }
		wordstart	{ delregion $w [$w index {insert wordstart}] insert }
		nextword	{ deletenextword $w insert }
		linestart	{ delregion $w [$w index {insert linestart}] insert }
		lineend		{ delregion $w insert [$w index {insert lineend}] }
		default		{ showinfo "Invalid command <delete $arg>." }
	}
}

proc command:selection {w arg text} {
	global	cutbuffer

	switch -exact $arg {
		delete		{ $w tag remove sel 1.0 end }
		copy		{ set cutbuffer [selection_get $w] }
		cut		{ set cutbuffer [selection_del $w] }
		paste		{ if {![catch {set t [selection get -displayof $w]}]} {command [list insert text $t]} }
		indent		{ selection_indent $w \t }
		reindent	{ selection_reindent $w \t }
		remark		{ selection_remark $w }
		delrem		{ selection_delremark $w }
		tag		{ selection_tag $w tag $text }
		untag		{ selection_tag $w untag $text }
		toggle		{ selection_tag $w toggle $text }
		send		{ selection_send $w $text }
		default		{ showinfo "Invalid command <selection $arg>." }
	}
}

proc command:search {w arg text} {
	global	search

	switch -exact $arg {
		panel		{ edit_search }
		search		{ search:findtext $text }
		forward		{ search:forward $text }
		backward	{ search:backward $text }
		replace		{ search:replacetext $text }
		replaceall	{ search:replacealltext $text }
		swapstr		{ search:swapstrings }
		exit		{ search:exit }
		searchstr	{ set search(search) $text }
		replacestr	{ set search(replace) $text }
		setexact	{ set search(type) {-exact} }
		setregexp	{ set search(type) {-regexp} }
		setcase		{ set search(case) {} }
		setnocase	{ set search(case) {-nocase} }
		setforward	{ set search(dir) {-forwards} }
		setbackward	{ set search(dir) {-backwards} }
		backup		{ search_backup }
		restore		{ search_restore }
		default		{ showinfo "Invalid command <search $arg>." }
	}
}

proc command:compile {w arg} {
	switch -exact $arg {
		make		{ compile_make }
		delerrors	{ compile:deleteerrors }
		acterror	{ compile:showacterror }
		preverror	{ compile:showpreverror }
		nexterror	{ compile:shownexterror }
		default		{ showinfo "Invalid command <compile $arg>." }
	}
}

proc command:file {w arg text} {
	switch -exact $arg {
		mode		{ file_set_mode $text }
		prev		{ file_prev }
		next		{ file_next }
		new		{ file_new }
		open		{ file_open $text }
		insert		{ file_insert $text }
		write		{ file_write $text }
		print		{ file_print }
		reopen		{ file_reopen }
		save		{ file_save }
		saveas		{ file_saveas }
		saveall		{ file_saveall }
		close		{ file_close }
		closeall	{ file_closeall }
		exit		{ file_exit }
		default		{ showinfo "Invalid command <file $arg>." }
	}
}

proc command:macro {w arg name} {
	global	macro

	switch -exact $arg {
		start		{ macro_start $name }
		stop		{ macro_stop }
		exec		{ macro_exec $name }
		wrap		{ macro_wrap }
		save		{ macros_save }
		load		{ macros_load }
		default		{ showinfo "Invalid command <macro $arg>." }
	}
}

proc command:rcs {w arg} {

	switch -exact $arg {
		new		{ rcs_new }
		chkin		{ rcs_chkin }
		chkout		{ rcs_chkout }
		default		{ showinfo "Invalid command <rcs $arg>." }
	}
}

proc command:project {w arg text} {
	global	project

	switch -exact $arg {
		open		{ project_open $text}
		saveas		{ project_saveas }
		save		{ project:save_file $project(name) }
		exit		{ project_close }
		add		{ project_add_file }
		delete		{ project_del_file }
		sort		{ set project(sort) $text; project_sort_new }
		default		{ showinfo "Invalid command <project $arg>." }
	}
}

proc command:arguments { argv } {
	if { $argv == "" } {
		set argv \(Noname\)
	}
	foreach arg $argv {
		new_file $arg {} {}
	}
}

proc command {line} {
	global	comcount cutbuffer macros macro fno

	incr comcount
	set w [twidgetbyno $fno]
	set com [lindex $line 0]
	set arg [lindex $line 1]
	if {[winfo exists $w] == 0} {
		if {$com != {file} && $com != {help} && $com != {project} && $com != {goto}} {
			return
		}
		if {$arg != {open} && $arg != {document} && $arg != {html} && $arg != {exit} && $arg != {new} && $arg != {fileno}} {
			return
		}
	}
	if {$macro(rec) && !$macro(play) && !$macro(wrap)} {
		if {$com != {macro}} {
			if {$com == {insert} && $arg == {text}} {
				append macro(last) [lindex $line 2]
			} else {
				if {$macro(last) != {}} {
					lappend macros($macro(name)) [list insert text $macro(last)]
					set macro(last) {}
				}
				lappend macros($macro(name)) $line
			}
		} else {
			if {$arg != {wrap}} {
				if {$macro(last) != {}} {
					lappend macros($macro(name)) [list insert text $macro(last)]
					set macro(last) {}
				}
				if {$arg == {exec}} {
					lappend macros($macro(name)) $line
				}
			}
		}
	}
	set macro(wrap) 0
	switch -exact $com {
		insert		{ command:insert $w $arg [lindex $line 2] }
		help		{ command:help $w $arg [lindex $line 2] }
		showhelp	{ command:showhelp $w $arg [lindex $line 2] }
		move		{ command:move $w $arg [lindex $line 2] }
		goto		{ command:goto $w $arg [lindex $line 2] }
		mark		{ command:mark $w $arg }
		delete		{ command:delete $w $arg }
		selection	{ command:selection $w $arg [lindex $line 2] }
		search		{ command:search $w $arg [lindex $line 2] }
		compile		{ command:compile $w $arg }
		file		{ command:file $w $arg [lindex $line 2] }
		macro		{ command:macro $w $arg [lindex $line 2] }
		rcs		{ command:rcs $w $arg }
		project		{ command:project $w $arg [lindex $line 2] }
		wssel		{ command:wssel $w $arg }
		tcl		{ uplevel #0 $arg }
		default		{ showinfo "Invalid command <$com>." }
	}
	if {$com != {file} || $arg != {close}} {
		if {[winfo exists $w] == 1} {
		showposition $w
		}
	}
}
# $Extended$File$Info$
#
