
proc compile:deleteerrors {} {
	global	compile files filename options

	set compile(errors) {}
	for {set i 0} {$i < $files} {incr i} {
		if {$filename($i) != ""} {
			set ew [twidgetbyno $i]
			foreach tag $options(compile.tags) {
				$ew tag remove $tag 1.0 end
			}
		}
	}
	update idletasks
}

proc compile:showerror {no} {
	global	compile info

	set nerr [llength $compile(errors)]
	if {$no >= $nerr} {
		set no [expr $nerr-1]
		if {$no < 0} return
	}
	set compile(acterr) $no
	set err [lindex $compile(errors) $no]
	if {[lindex $err 2] == {warn}} {
		set str {WARNING}
	} else {
		set str {ERROR}
	}
	set fn [nobyname [lindex $err 0]]
	set ew [twidgetbyno $fn]
	set lno [lindex $err 1]
	.nb raise $fn
	$ew mark set insert $lno.0
	$ew see $lno.0
	set info(text) "$str ([expr $compile(acterr)+1]/$nerr): [lindex $err 3]"
	incr info(no)
	update idletasks
	focus $ew
}

proc compile:showacterror {} {
	global	compile

	compile:showerror $compile(acterr)
}

proc compile:shownexterror {} {
	global compile

	set n [llength $compile(errors)]
	if {$compile(acterr) < $n-1} {
		compile:showerror [incr compile(acterr)]
	}
}

proc compile:showpreverror {} {
	global compile

	set i [llength $compile(errors)]
	if {$compile(acterr) > 0} {
		compile:showerror [incr compile(acterr) -1]
	}
}

proc compile_command {} {
	global	compilecmd tkPriv program

	set oldcmd $compilecmd
	set w .compilecmd
	catch {destroy $w}
	toplevel $w
	wm title $w "$program(name): compile command"
	wm iconname $w "$program(name)"

	frame $w.t -bd 1 -relief raised
	label $w.t.l -text "Command:" -width 10 -anchor e
	entry $w.t.e -textvariable compilecmd -width 20 -bd 2 -relief sunken
	pack $w.t.l $w.t.e -side left -padx 5 -pady 5 -fill both -expand yes
	frame $w.b -bd 1 -relief raised
	button $w.b.ok -text Ok -command {set tkPriv(button) 1}
	button $w.b.cancel -text Cancel -command {set tkPriv(button) 0}
	pack $w.b.ok $w.b.cancel -side left -padx 10 -pady 5 -fill both -expand yes
	pack $w.t $w.b -side top -fill x -expand yes
	update idletasks
	bind $w <Destroy> {set tkPriv(button) 0}
	bind $w <Return> "
		$w.b.ok configure -state active -relief sunken
		update idletasks
		after 100
		set tkPriv(button) 1
	"
	bind $w <Escape> "
		$w.b.cancel configure -state active -relief sunken
		update idletasks
		after 100
		set tkPriv(button) 0
	"
	set oldfocus [focus]
	focus $w.t.e
	tkwait variable tkPriv(button)
	catch {focus $oldfocus}
	catch {
		bind $w <Destroy> {}
		destroy $w
	}
	if {!$tkPriv(button)} {
		set compilecmd $oldcmd
	}
	
}

proc compile_make {} {
	global	files filename compile info fno program fmode

	if {![file exists Makefile] && ![file exists makefile]} {
		showinfo "No 'Makefile' or 'makefile' found. <$fmode>"
		return
	}
	if {![catch {exec make -q}]} {
		set compile(errors) {}
		showinfo "Target is up to date."
		return
	}
	compile:deleteerrors
	set w .compilewin
	if {!$compile(done)} {
		$w.t delete 1.0 end
	} else {
		set compile(done) 0
		catch {destroy $w}
		toplevel $w
		wm title $w "$program(name): compile make"
		wm iconname $w "$program(name)"
	
		frame $w.o -bd 1 -relief raised
		text $w.t -width 60 -height 20 -wrap none -state normal -exportselection yes -relief sunken -bd 2\
				-yscrollcommand "$w.y set" -xscrollcommand "$w.x set"
		scrollbar $w.y -command "$w.t yview"
		scrollbar $w.x -command "$w.t xview" -orient h
		pack $w.y -in $w.o -side right -fill y
		pack $w.x -in $w.o -side bottom -fill x
		pack $w.t -in $w.o -side top -expand yes -fill both
		frame $w.b -bd 1 -relief raised
		button $w.make -text "Make" -command {command {compile make}}
		button $w.prev -text "Prev error" -command {command {compile preverror}}
		button $w.next -text "Next error" -command {command {compile nexterror}}
		button $w.act -text "Actual error" -command {command {compile acterror}}
		button $w.del -text "Delete errors" -command {command {compile delerrors}}
		button $w.done -text "Done" -command {set compile(done) 1}
		pack $w.make $w.prev $w.next $w.act $w.del $w.done -in $w.b -side left -fill x
		pack $w.o -side top -fill both -expand yes
		pack $w.b -side bottom -fill x
		bind $w <Destroy> {set compile(done) 1}
		bind $w <Return> "
			$w.done configure -state active -relief sunken
			update idletasks
			after 100
			set compile(done) 1
		"
		bind $w <Escape> "
			$w.done configure -state active -relief sunken
			update idletasks
			after 100
			set compile(done) 1
		"
	}
	set lno 0
	set message {}
	set omessage $message
	set olno $lno
	set mdir {}
	showinfo "Compiling ...(can take a while!)"
	set f [open "| make -s -w |& cat"]
	while {[set i [gets $f line]] != -1} {
		$w.t insert insert "$line\n"
		$w.t see insert
		update idletasks
		set ml [split $line {:}]
		set name [lindex $ml 0]
		set lno [lindex $ml 1]
		set message [string trimleft [lindex $ml 2]]
		if {[string trim $message] == {warning}} {
			set type warn
			set message [string trimleft [lindex $ml 3]]
		} else {
			set type err
		}
		foreach m [lrange $ml [expr 3+("$type"=={warn})] end] {
			append message ":$m"
		}
		if {$name == {make} && [lindex $lno 0] == {Entering} && [lindex $lno 1] == {directory}} {
			set s [lindex $lno 2]
			set mdir [string range $s 1 [expr [string length $s]-2]]
		}
		if {[file exists $mdir/$name] && [regexp {[0-9]+} $lno] && $lno != $olno} {
			file_goto_named $mdir/$name
			set ew [twidgetbyno $fno]
			$ew tag add $type $lno.0 "$lno.0+1l"
			lappend compile(errors) [list $name $lno $type $message]
			$ew see $lno.0
			update idletasks
		}
		set omessage $message
		set olno $lno
	}
	set compile_error [catch {close $f}]
	if {$compile_error} {
		showinfo "Compiling ... failed. ([llength $compile(errors)] errors found)"
	} else {
		showinfo "Compiling ... done. ([llength $compile(errors)] errors found)"
	}
	compile:showerror 0
	tkwait variable compile(done)
	catch {
		bind $w <Destroy> {}
		destroy $w
	}
}

proc compile_makeclean {} {
	global fmod

	if {![file exists Makefile] && ![file exists makefile]} {
		showinfo "No 'Makefile' or 'makefile' found. $fmode"
		return
	}
	showinfo "Make clean ..."
	catch {exec make clean}
	showinfo "Make clean ... done."
}
