
proc edit_search {} {
	global	fno program search

	if {$search(panel) != {}} {
		focus $search(panel).searchstr
		return
	}

	set w .sb
	toggle_bar $w .nb global.true
	
	set search(panel) $w
	tkwait visibility $w.searchstr
	focus $w.searchstr
}

proc search:swapstrings {} {
	global	search

	set search(tmp) $search(search)
	set search(search) $search(replace)
	set search(replace) $search(tmp)
}

proc search:findtext {text} {
	global	search fno

	if {$text != {}} {
		set search(search) $text
	}
	if {$search(search) == {}} {
		showinfo "Search string is empty!"
		return
	}
	showinfo "Searching ..."
	set w [twidgetbyno $fno]
	set c [lindex [findtext $w] 1]
	if {$c > 0} {
		showinfo "Searching ... done."
	} else {
		showinfo "Pattern '$search(search)' not found."
	}
	showposition $w
}

proc search:replacetext {text} {
	global	search fno

	if {$text != {}} {
		set search(search) $text
	}
	if {$search(search) == {}} {
		showinfo "Search string is empty!"
		return
	}
	showinfo "Replacing & Searching ..."
	set w [twidgetbyno $fno]
	set c [lindex [replacefindtext $w] 1]
	if {$c > 0} {
		showinfo "Replacing & Searching ... done."
	} else {
		showinfo "Pattern '$search(search)' not found."
	}
	showposition $w
}

proc search:replacealltext {text} {
	global	search fno

	if {$text != {}} {
		set search(search) $text
	}
	if {$search(search) == {}} {
		showinfo "Search string is empty!"
		return
	}
	showinfo "Replacing & Searching ALL ..."
	set w [twidgetbyno $fno]
	replacefindalltext $w
	showposition $w
	showinfo "Replacing & Searching ALL ... done."
}

proc search:exit {} {
	global	search

	if {[lsearch [winfo children .] .sb] != -1} {
		focus [twidgetbyno [getraised]]
		set search(panel) {}
		pack forget .sb
	}
}

proc search:forward {text} {
	global	search

	set search(dir)	{-forwards}
	search:findtext $text
}

proc search:backward {text} {
	global	search

	set search(dir)	{-backwards}
	search:findtext $text
}

proc special_char_panel {w} {
	global	program tkPriv options

	set tkPriv(special_char_panel) $w
	set tkPriv(button) 0
	catch {destroy $w}
	toplevel $w
	wm title $w "$program(name): special characters"
	wm iconname $w "$program(name)"

	frame $w.t -bd 1 -relief raised
	for {set row 0} {$row < 16} {incr row} {
		frame $w.t.r$row
		for {set col 0} {$col < 16} {incr col} {
			set ch [format "%c" [expr $row*16+$col]]
			eval button $w.t.r$row.c$col -font $options(tag.default.font) -command \{command \"insert char [expr $row*16+$col]\"\}
			$w.t.r$row.c$col configure -text $ch
			pack $w.t.r$row.c$col -side left -expand yes -fill both
		}
		pack $w.t.r$row -side top -expand yes -fill both
	}
	frame $w.b -bd 1 -relief raised
	button $w.b.cancel -text Cancel -command {set tkPriv(button) 0}
	pack $w.b.cancel -side left -padx 10 -pady 5 -fill x -expand yes
	pack $w.t $w.b -side top -fill x -expand yes
	update idletasks
	bind $w <Destroy> {set tkPriv(button) -1}
	bind $w <Return> "
		$w.b.cancel configure -state active -relief sunken
		update idletasks
		after 100
		set tkPriv(button) 0
	"
	bind $w <Escape> "
		$w.b.cancel configure -state active -relief sunken
		update idletasks
		after 100
		set tkPriv(button) 0
	"
	set oldfocus [focus]
	focus $w.b.cancel
	tkwait variable tkPriv(button)
	catch {focus $oldfocus}
	catch {
		bind $w <Destroy> {}
		destroy $w
	}
	return $tkPriv(button)
}

proc selection_tag_panel {w} {
	global	options tagstate fno program fileopts fileefi

	set t [twidgetbyno $fno]
	set tags $options(user.tags)
	foreach tag $fileefi($fno.tags) {
		if {[lsearch $tags $tag] == -1} {
			lappend tags $tag
		}
	}
	set selrange {}
	if {[selection_exists $t]} {
		set selrange [selection_range $t]
	}
	foreach tag $tags {
		set ranges [$t tag ranges $tag]
		set l [llength $ranges]
		set state -1
		if {$selrange != {}} {
			for {set i 0} {$i < $l} {incr i} {
				set b [lindex $ranges $i]
				incr i
				set e [lindex $ranges $i]
				if {[$t compare $b < [lindex $selrange 1]] && [$t compare $e > [lindex $selrange 0]]} {
					if {[$t compare $b <= [lindex $selrange 0]]} {
						if {[$t compare $e >= [lindex $selrange 1]]} {
							set state 1
						} else {
							set state 0
						}
					} else {
						if {[$t compare $e >= [lindex $selrange 1]]} {
							set state 1
						} else {
							set state 0
						}
					}
				}
			}
		}
		set tagstate($tag) $state
	}
	set tkPriv(tag_panel) $w
	set tkPriv(button) 0
	catch {destroy $w}
	toplevel $w
	wm title $w "$program(name): tag selection"
	wm iconname $w "$program(name)"

	frame $w.t -bd 1 -relief raised
	foreach tag $tags {
		frame $w.t.$tag
		if {[info exists options(tag.$tag.info)]} {
			label $w.t.$tag.l -text "$options(tag.$tag.info):" -bd 2 -relief ridge
		} elseif {[info exists fileefi($fno.tag.$tag.info)]} {
			label $w.t.$tag.l -text "$fileefi($fno.tag.$tag.info):" -bd 2 -relief ridge
		} else {
			label $w.t.$tag.l -text "$tag:" -bd 2 -relief ridge
		}
		radiobutton $w.t.$tag.untag -text untag -variable tagstate($tag) -value -1 -command [list command [list selection untag $tag]]
		radiobutton $w.t.$tag.part -text partially -variable tagstate($tag) -value 0 -state disabled
		radiobutton $w.t.$tag.tag -text tag -variable tagstate($tag) -value 1 -command [list command [list selection tag $tag]]
		pack $w.t.$tag.l -side left
		pack $w.t.$tag.tag $w.t.$tag.part $w.t.$tag.untag -side right
		pack $w.t.$tag -side top -expand yes -fill x
	}
	frame $w.b -bd 1 -relief raised
	button $w.b.ok -text Done -command {set tkPriv(button) 0}
	pack $w.b.ok -side left -padx 10 -pady 5 -fill x -expand yes
	pack $w.t $w.b -side top -fill x -expand yes
	update idletasks
	bind $w <Destroy> {set tkPriv(button) 0}
	bind $w <Return> "
		$w.b.ok configure -state active -relief sunken
		update idletasks
		after 100
		set tkPriv(button) 0
	"
	bind $w <Escape> "
		$w.b.ok configure -state active -relief sunken
		update idletasks
		after 100
		set tkPriv(button) 0
	"
	set oldfocus [focus]
	focus $w.b.ok
	tkwait variable tkPriv(button)
	catch {focus $oldfocus}
	catch {
		bind $w <Destroy> {}
		destroy $w
	}
	return $tkPriv(button)
}
