
proc file_dialog_setpattern {pattern} {
	global	filedialog

	set l {}
	foreach p $pattern {
		if {$p == {--}} break
		lappend l $p
	}
	$filedialog subwidget fsbox subwidget file config -value $l
	return $pattern
}

proc file_dialog {title type command file} {
	global	filedialog program fileoldpattern fileext_support

		if {$filedialog == {}} {
			set filedialog [tix filedialog tixExFileSelectDialog]
			$filedialog subwidget fsbox config -filetypes $fileext_support
			$filedialog subwidget fsbox subwidget types config -validatecmd file_dialog_setpattern
			$filedialog subwidget fsbox subwidget types pick [expr [llength $fileext_support]-1]
		}

		if {$fileoldpattern != {}} {
			$filedialog subwidget fsbox subwidget types pick $fileoldpattern
			set fileoldpattern {}
		}

		$filedialog config -command $command -title "$program(name): $title"
		if {$type != {}} {
			set fileoldpattern [$filedialog subwidget fsbox subwidget types subwidget listbox index anchor]
		}
		switch $type {
			prj	{
				$filedialog subwidget fsbox subwidget types pick [expr [llength $fileext_support]-2]
			}
		}
		$filedialog subwidget fsbox filter
		$filedialog popup
}

proc setmode {no type} {
	global	modes mode fno fmode options

	if {[lsearch $modes $type] == -1} { return }
	if {$no == $fno} {set fmode $type}
	set mode($no) $type
	set f [twidgetbyno $no]
	foreach tag $options(color.tags) {
		$f tag remove $tag 1.0 end
	}
	if {$mode($no) == {}} {
		showinfo "Remove colorisation."
	} else {
		set s $mode($no)
		showinfo "Colorising ($s-mode) ... (can take a while!)"
		colorize $f 1.0 end
		newremark $f
		showinfo "Colorising ($s-mode) ... done."
	}
	return $type
}

proc file_set_mode {args} {
	global	fno files mode fmode

	set w [twidgetbyno $fno]
	if {[winfo exists $w] == 0} return

	if {$fno > $files} return
	if {[llength $args]} {
		set fmode [lindex $args 0]
		if {$mode($fno) == [lindex $args 0]} return
		setmode $fno [lindex $args 0]
	} elseif {$mode($fno) != {}} {
		set mode($fno) {}
		setmode $fno {}
	}
}

proc file_opt_set {args} {
	global	fileopts fno

	set opt [lindex $args 0]
	set fileopts($fno.$opt) [uplevel #0 set f$opt]
}

proc file_set_opts {args} {
	global	fileopts fopts fno

	set fopts $args
	foreach opt [array names fileopts $fno.*] {
		if {[lsearch $args [extension $opt] != -1} {
			set val 1
		} else {
			set val 0
		}
		set fileopts($opt) $val
	}
}

proc modebyext {name} {
	global	modes mode fileext

	set type {}
	foreach ftype $modes {
		if {[lsearch $fileext($ftype) [string tolower [extension $name]]] != -1} {
			set type $ftype
		}
	}
	return $type
}

proc init_file {no} {
	global	fno undo undovars withundo fileefi

	setmodified $no 0
	init_undo $no
	set withundo		1
	set fileefi($no.tags)	{}
}

proc get_file_options {no} {
	global	fileopts

	set opts {}
	foreach opt [array names fileopts $no.*] {
		lappend opts [list [extension $opt] $fileopts($opt)]
	}
	return $opts
}

proc set_file_options {no opts} {
	global	options fileopts fileefi

	set fileefi($no.tags) {}
	foreach opt [array names options file.*] {
		set fileopts($no[file extension $opt]) $options($opt)
	}
	foreach opt $opts {
		set fileopts($no.[lindex $opt 0]) [lindex $opt 1]
	}
}

proc addpos {pos add} {

	scan $pos {%d.%d} y x
	scan $add {%d.%d} yd xd
	if {$y == 1} {
		incr x $xd
	}
	incr y [expr $yd-1]
	return $y.$x
}

proc subpos {pos add} {

	scan $pos {%d.%d} y x
	scan $add {%d.%d} yd xd
	if {$y == $yd} {
		incr x -$xd
	}
	incr y [expr -$yd+1]
	return $y.$x
}

proc load_file_efi {w f no add} {
	global	efi mode fileopts options fileefi

	set fileopts($no.useefi) 1
	while {[gets $f l] >= 0} {
		set l [string range $l [string length [lindex $efi($mode($no)) 2]] end]
		set ll [llength $l]
		switch -exact [lindex $l 0] {
			bookmarks {
				for {set i 1} {$i < $ll} {incr i} {
					set n [lindex $l $i]
					incr i
					set p [addpos [lindex $l $i] $add]
					bookmark_set $w $n $p
				}
			}
			tagdefs {
				set tag [lindex $l 1]
				if {[lsearch $fileefi($no.tags) $tag] == -1} {
					lappend fileefi($no.tags) $tag
					set fileefi($no.tag.$tag.info) $tag
					for {set i 2} {$i < $ll} {incr i} {
						set t [lindex $l $i]
						incr i
						set v [lindex $l $i]
						set fileefi($no.tag.$tag.$t) $v
					}
				}
			}
			tags {
				if {[lsearch $options(user.tags) [lindex $l 1]] != -1 || [lsearch $fileefi($no.tags) [lindex $l 1]] != -1} {
					for {set i 2} {$i < $ll} {incr i} {
						set b [addpos [lindex $l $i] $add]
						incr i
						set e [addpos [lindex $l $i] $add]
						$w tag add [lindex $l 1] $b $e
					}
				}
			}
		}
	}
}

proc read_file {w name mode no} {
	global	efi

	set add [$w index insert]
	set text {}
	set f [open $name r]
	set efs "[lindex $efi($mode) 0] $efi(start)"
	deletetags $w
	tagcolors $w $no
	set justefi 0
	if {[gets $f l] >=0} {
		if {"$efs" == "$l"} {
	 		puttext $w $text
		 	set text {}
		 	load_file_efi $w $f $no $add
			set justefi 1 
		} else {
			set justefi 0
		}
	
	 }
	if {$justefi != 1} {
		append text "$l"
		while {[gets $f l] >= 0} {
			if {"$efs" == "$l"} {
				puttext $w $text
				set text {}
				load_file_efi $w $f $no $add
				break;
		}
		append text "\n$l"
	}
	}
	if {$text != {}} {
		puttext $w $text
	}
	tagcolors $w $no
	close $f
}

proc load_file {no name opts coms} {
	global	mode info files filename options mode

	set w [twidgetbyno $no]
	if {$name == {} || $name == {(Noname)}} {
		file_open
		return
	}
	showinfo "Loading file <[file tail $name]> ..."
	if {[file readable $name]} {
		$w delete 1.0 end
		set filename($no) [completename $name]
		set_file_options $no $opts
		set mode($no) [modebyext $name]
		init_file $no
		.nb pageconfigure $no -label [file tail $name]
		deletetags $w
		read_file $w $name $mode($no) $no
		setmodified $no 0
		init_undo $no
		setcursor $w 1.0
		eval $coms
		setmode $no $mode($no)
		getraiseno
		showinfo "Loading file <[file tail $name]> ... done."
	} else {
		showinfo "Unable to load file <[file tail $name]>!"
	}
}

proc posinrange {w first last pos} {

	return [expr [$w compare $pos >= $first] && [$w compare $pos <= $last]]
}

proc rangeinrange {w first last i1 i2} {

	if {[$w compare $i2 <= $first] || [$w compare $i1 >= $last]} {
		return {}
	}
	if {[$w compare $i1 < $first]} {
		set i1 $first
	}
	if {[$w compare $i2 > $last]} {
		set i2 $last
	}
	return "[subpos $i1 $first] [subpos $i2 $first]"
}

proc write_file_efi {w f no first last} {
	global	fileopts efi mode options fileefi

	if {$fileopts($no.useefi)} {
		set efistart [lindex $efi($mode($no)) 0]
		set efiline [lindex $efi($mode($no)) 1]
		set efiend [lindex $efi($mode($no)) 2]
		puts $f "$efistart $efi(start)"
		set bmstr "$efiline bookmarks "
		set bmc 0
		foreach bm [$w mark names] {
			if {[regexp {bm.*} $bm]} {
				if {[bookmark_defined $w [string range $bm 2 end]]} {
					if {[posinrange $w $first $last [set idx [$w index $bm]]]} {
						incr bmc
						lappend bmstr [string range $bm 2 end] [subpos $idx $first]
					}
				}
			}
		}
		if {$bmc} {
			puts $f $bmstr
		}
		foreach tag $fileefi($no.tags) {
			set defs {}
			foreach t [array names fileefi $no.tag.$tag.*] {
				lappend defs [extension $t] $fileefi($t)
			}
			puts $f "$efiline tagdefs $tag $defs"
		}
		foreach tag "$options(user.tags) $fileefi($no.tags)" {
			if {[set ranges [$w tag ranges $tag]] != {}} {
				set rangelist {}
				set ll [llength $ranges]
				for {set i 0} {$i < $ll} {incr i} {
					set i1 [lindex $ranges $i]
					incr i
					set i2 [lindex $ranges $i]
					if {[set r [rangeinrange $w $first $last $i1 $i2]] != {}} {
						lappend rangelist [lindex $r 0] [lindex $r 1]
					}
				}
				if {$rangelist != {}} {
					puts $f "$efiline tags $tag $rangelist"
				}
			}
		}
		puts $f "$efiend"
	}
}

proc write_file {no name} {
	global	program options

	if {$name == {} || $name == {(Noname)}} {
		file_saveas
		return
	}
	showinfo "Saving file <[file tail $name]> ..."
	if {[file exists $name]} {
		if {[catch {exec cp -f $name $name$options(backupsuffix)}]} {
			if {![tk_dialog .fcd "$program(name): save" "Failed to make a backup file. Save anyway?" warning 0 "No" "Save!"]} {
				showinfo "Saving file <[file tail $name]> ... aborted."
				return
			}
		}
	} else {
		catch {exec touch $name}
	}
	if {[file writable $name]} {
		set w [twidgetbyno $no]
		set f [open $name w]
		puts -nonewline $f [$w get 1.0 end]
		write_file_efi $w $f $no 1.0 end
		close $f
		showinfo "Saving file <[file tail $name]> ... done."
	} else {
		showinfo "Unable to save <[file tail $name]>!"
	}
}

proc new_file {name opts coms} {
	global	files filename font color back options mode fileopts

	if {$name == {}} {
		return
	} else {
		if {$name != {(Noname)}} {
			for {set i 0} {$i < $files} {incr i} {
				if {$filename($i) == $name} {
					showinfo "File [file tail $name] allready exists."
					return
				}
			}
		}
	}
	set mode($files) {}
	foreach opt [array names options file.*] {
		set fileopts($files[file extension $opt]) $options($opt)
	}
	.nb add $files -label [file tail $name] -raisecmd {getraiseno}
	set f [widgetbyno $files]
	text $f.t -exportselection yes -relief sunken -bd 2 -yscrollcommand "$f.ysc set" -xscrollcommand "$f.xsc set"
	foreach option [array names options tag.default.*] {
		set config [extension $option]
		if {$config != {del} && $config != {info}} {
			catch {$f.t configure -$config $options($option)}
		}
	}
	scrollbar $f.ysc -command "$f.t yview" -takefocus 0
	scrollbar $f.xsc -command "$f.t xview" -takefocus 0 -orient h
	pack $f.ysc -side right -fill y
	pack $f.xsc -side bottom -fill x
	pack $f.t -side top -expand yes -fill both
	.nb raise $files
	if {$name != {} && $name != {(Noname)} && [file readable $name]} {
		load_file $files $name $opts $coms
	} else {
		set_file_options $files {}
		set filename($files) $name
		init_file $files
		if {$name != {} && $name != {(Noname)}} {
			setmode $files [modebyext $name]
		} else {
			setmode $files {}
		}
		tagcolors $f.t $files
		getraiseno
	}
	update idletasks
	focus $f.t
	incr files
}

proc file_new {} {
	new_file {(Noname)} {} {}
}

proc file_open {file} {

	file_dialog {file open} {} {file:open_file} $file
}

proc file:open_file {file} {
	global	filename mode

	if {$file != {}} {
		new_file $file {} {}
	}
	setfocus
}

proc file_insert {file} {

	file_dialog {file insert} {} {file:insert_file} $file
}


proc file:insert_file {file} {
	global	filename mode fno options

	if {$file != {}} {
		if {[file readable $file]} {
			showinfo "Reading file <[file tail $file]> ..."
			set t {}
			set w [twidgetbyno $fno]
			set p [$w index insert]
			read_file $w $file [modebyext $file] $fno
			puttext $w $t
			$w tag remove sel 1.0 end
			if {$options(global.preferwssel)} {
				wssel_begin $w $p
				wssel_end $w insert
			} else {
				$w tag add sel $p insert
			}
			update idletasks
			showinfo "Reading file <[file tail $file]> ... done."
		} else {
			showinfo "File <[file tail $file]> is not readable!"
		}
	}
	setfocus
}

proc file_write {file} {
	global	fno writebuffer writestart writeend

	set writebuffer {}
	set w [twidgetbyno $fno]
	if {![selection_exists $w]} {
		showinfo "No selection to write to file."
		return
	} else {
		set selrange [selection_range $w]
		set writestart [lindex $selrange 0]
		set writeend [lindex $selrange 1]  
		set writebuffer [[selection_w $w] get $writestart $writeend]
	}
	file_dialog {write to file} {} {file:write_file} $file
}

proc file:write_file {file} {
	global	filename mode fno writebuffer writestart writeend

	set w [twidgetbyno $fno]
	if {![selection_exists $w] && $writebuffer == {}} {
		showinfo "No selection to write to file."
		return
	}
	if {$file != {}} {
		showinfo "Writing to file <[file tail $file]> ..."
		exec touch $file
		if {[file writable $file]} {
			set f [open $file w]
			if {$writebuffer == {}} {
				set selrange [selection_range $w]
				set writestart [lindex $selrange 0]
				set writeend [lindex $selrange 1]
				set writebuffer [[selection_w $w] get $writestart $writeend]
			}
			puts $f $writebuffer
			write_file_efi $w $f $fno $writestart $writeend
			close $f
			showinfo "Writing to file <[file tail $file]> ... done."
		} else {
			showinfo "Unable to write to <[file tail $file]>!"
		}
	}
	setfocus
	set writebuffer {}
}

proc save_file {no} {
	global	filename fno

	if {$no == {}} {
		set fn $fno
	} else {
		set fn $no
	}
	if {$filename($fn) != {}} {
		if {[file tail $filename($fn)] == {(Noname)}} {
			saveas_file $fn
		} else {
			write_file $fn $filename($fn)
			setmodified $fn 0
		}
	}
}

proc file_save {} {
	global	fno

	save_file $fno
}

proc saveas_file {no} {
	global	fsaveas filename

	set fsaveas $no
	file_dialog {file saveas} {} {file:save_file} $filename($no)
}

proc file_saveas {} {
	global	fno

	saveas_file $fno
}

proc file:save_file {file} {
	global	filename fsaveas mode

	if {$fsaveas == {}} {
		set fn [getraised]
	} else {
		set fn $fsaveas
	}
	if {$file != {}} {
		set filename($fn) [completename $file]
		.nb pageconfigure $fn -label [file tail $filename($fn)]
		setmode $fn [modebyext $file]
		write_file $fn $file
		setmodified $fn 0
		getraiseno
	}
	setfocus
}

proc file_saveall {} {
	global	files filename

	for {set i 0} {$i < $files} {incr i} {
		if {$filename($i) != {}} {
			save_file $i
		}
	}
}

proc file_free {no} {
	global	undo fileefi fileopts

	foreach name [array names undo $no.*] {
		unset undo($name)
	}
}

proc file_exit {} {
	global	program modified files project filename

	set modi 0
	set open 0
	for {set i 0} {$i < $files} {incr i} {
		incr modi $modified($i)
		if {[string trim $filename($i)] != {}} {
			incr open
		}
	}

	if {$modi} {
		if {[tk_dialog .fed "$program(name): exit" "There are modified texts. Do you really want to exit?" warning 0 "No" "Exit!"]} {
			set modi 0
		}
	}

	if {$project(name) != {} && $open != 0} {
		project_save
	}

	if {!$modi} {
		exit
	}
}

proc file_closed {fn} {
	global	program modified filename files project

		if {$modified($fn)} {
			if {[tk_dialog .fcd "$program(name): close" "The text has been modified. Do you really want to close?" warning 0 "No" "Close!"]} {
				setmodified $fn 0
			}
		}

		if {!$modified($fn)} {
			set filename($fn) {}
			wm title . "$program(name): empty"
			file_free $fn
			setmodified $fn 0
			.nb delete $fn
			update idletasks
			set fn [getraised]


		}
}

proc file_close {} {
	global	program modified filename files project

	set fn [getraised]
	if {$fn != {}} {

		file_closed $fn
	}
}

proc file_closeall {} {
	global	files filename

	for {set i 0} {$i < $files} {incr i} {
		if {$filename($i) != {}} {
			file_closed $i
		}
	}
}

proc file_reopen {} {
	global	program modified filename files fno

	if {$modified($fno)} {
		if {[tk_dialog .fcd "$program(name): reopen" "The text has been modified. Do you really want to reopen?" warning 0 "No" "Reopen!"]} {
			setmodified $fno 0
		}
	}

	if {!$modified($fno)} {
		load_file $fno $filename($fno) {} {}
	}
}

proc file_prev {} {
	global	filename files

	set ex 0
	set fn [getraised]
	while {!$ex} {
		if {$fn > 0} {
			incr fn -1
		} else {
			set fn [expr $files-1]
		}
		if {$filename($fn) != {}} {set ex 1}
	}
	.nb raise $fn
	showinfo "Switch to [file tail $filename($fn)]."
	focus [twidgetbyno $fn]
}

proc file_goto_named {name} {
	global	filename filenamelist files project tkPriv

	if {[string trim $name] == {}} {
		if {[input_panel .fgn {goto file} {Filename} 20 {} $filenamelist]} {
			if {[string trim $tkPriv(input)] != {}} {
				command [list goto file $tkPriv(input)]
			}
		}
		return
	}
	if {[lsearch $filenamelist [string trim $name]] == -1} {
		lappend filenamelist [string trim $name]
	}
	for {set i 0} {$i < $files} {incr i} {
		if {[file tail $filename($i)] == [file tail $name]} {
			.nb raise $i
			showinfo "Switch to [file tail $filename($i)]."
			focus [twidgetbyno $i]
			return
		}
	}
	showinfo "No opened file <[file tail $name]> in editor!"
	after 500
	showinfo "Opening new file <[file tail $name]> ..."
	set opts {}
	set coms {}
	foreach entry $project(files) {
		if {[file tail [lindex $entry 0]] == [file tail $name]} {
			set opts [lindex $entry 4]
			set mode [lindex $entry 2]
			set pos [lindex $entry 1]
			set coms "set mode(\$no) $mode; setcursor \$w $pos"
		}
	}
	new_file $name $opts $coms
	showinfo "Opening new file <[file tail $name]> ... done."
}

proc file_next {} {
	global	filename files

	set ex 0
	set fn [getraised]
	while {!$ex} {
		if {$fn < $files-1} {
			incr fn
		} else {
			set fn 0
		}
		if {$filename($fn) != {}} {set ex 1}
	}
	.nb raise $fn
	showinfo "Switch to [file tail $filename($fn)]."
	focus [twidgetbyno $fn]
}

proc file_printer_select {} {
	global	tkPriv options

	if {[input_panel .fps {options printer} {Printer name} 15 $options(printer) {}]} {
		if {[string trim $tkPriv(input)] != {}} {
			set options(printer) [string trim $tkPriv(input)]
		}
	}
}

proc file_print {} {
	global	options filename fno modified program

	set t [[twidgetbyno $fno] get 1.0 end]
	set name [file tail $filename($fno)]
	if {$name == "(Noname)"} {
		set name Noname
	}
	set xtmp "/tmp/eddi.[pid]x.$name"
	set ftmp "/tmp/eddi.[pid]f.$name"
	if {$options(global.printa2ps)} {
		set f [open $xtmp w]
	} else {
		set f [open $ftmp w]
	}
	puts -nonewline $f $t
	close $f
	if {$options(global.printa2ps)} {
		showinfo {Filter file with a2ps ...}
		if {![catch {exec a2ps -q --output=- --center-title=[file tail $filename($fno)] --footer=$filename($fno) $xtmp > $ftmp}]} {
			showinfo {Filter file with a2ps ... done.}
		} else {
			showinfo {Filter file with a2ps ... failed.}
			return
		}
			
	}
	showinfo "Send file to printer <$options(printer)> ..."
	if {[string trim $options(printer)] != {}} {
		set printer "-P$options(printer)"
	} else {
		set printer {}
	}
	if {![catch {exec lpr -T $filename($fno) $printer < $ftmp}]} {
		showinfo "Send file to <$options(printer)> ... done."
	} else {
		showinfo "Send file to <$options(printer)> ... failed."
	}
	catch {exec rm -f /tmp/eddi.[pid]?.$name}
}

