
proc project_del_menu {} {
	.m.project.m delete 12 end
}

proc project_make_menu {} {
	global project

	project_del_menu
	.m.project.m add separator
	set chstr "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
	for {set i 0} {$i < [llength $project(files)]} {incr i} {
		set entry [lindex $project(files) $i]
		set ch [string range $chstr $i $i]
		.m.project.m add command -command "command \"goto fileno $i\"" -underline 0\
			-label "$ch: [file tail [lindex $entry 0]]" -accelerator "C-p [string tolower $ch]"
	}
}

proc project_comp_file {name num} {
	global project

	if {$num >= [llength $project(files)]} {
		return 0
	} elseif {$project(sort) == {no} && $num < [llength $project(files)]} {
		return 1
	}
	set lname [lindex [lindex $project(files) $num] 0]
	switch -exact $project(sort) {
		name	{
			return [string compare [file tail $name] [file tail $lname]]
		}
		ext	{
			switch -exact -- [set comp [string compare [file extension $name] [file extension $lname]]] {
				0	{
					return [string compare [file rootname [file tail $name]] [file rootname [file tail $lname]]]
				}
				default	{
					return $comp
				}
			}
		}
	}
}

proc project_add_entry {entry} {
	global project

	set i 0
	while {[project_comp_file [lindex $entry 0] $i]>0} {
		incr i
	}
	set project(files) [linsert $project(files) $i $entry]
}

proc project_new {file} {
	global project files filename modified argv0 program

	for {set i 0} {$i < $files} {incr i} {
		if {$modified($i) && $filename($i) != {}} {
			tk_dialog .fed "$program(name): new project" "There are modified files! - Save all before opening new project." warning 0 Ok
			return
		}
	}
	if {$project(name) != {} && !$project(saved)} {
		if {![tk_dialog .fed "$program(name): new project" "The project file <$project(name)> is unsaved. Do you want to discard it?" warning 0 "No" "Discard!"]} {
			return
		}
	}
	showinfo "Opening new project <[file tail $file]> ..."
	wm command . [list $argv0 [completename $file]]
	set oldfiles $files
	if {![file readable $file]} {
		showinfo "No project <[file tail $file]> found."
		return
	}
	set f [open $file r]
	set project(saved) 1
	set project(name) [completename [file rootname $file]]
	set project(info) [file tail $project(name)]
	set dir [file dirname [completename $file]]
	cd $dir
	uplevel #0 cd $dir
	set project(files) {}
	set project(sort) no
	set pfiles {}
	set active {}
	while {[gets $f line] > -1} {
		switch -exact [lindex $line 0] {
			file:	{
				lappend pfiles [list [lindex $line 1] [lindex $line 2] [lindex $line 3] [lindex $line 4] [lrange $line 5 end]]
			}
			sort:	{
				set project(sort) [lindex $line 1]
			}
			geometry: {
				catch {wm geometry . [lindex $line 1]}
			}
			active: {
				set active [lindex $line 1]
			}
		}
	}
	close $f
	foreach entry $pfiles {
		project_add_entry $entry
	}
	set nfiles 0
	foreach entry $project(files) {
		set n -1
		for {set i $nfiles} {$i < $files} {incr i} {
			if {$filename($i) != {}} {
				set n $i
				break;
			}
		}
		if {[lindex $entry 3]} {
			set fname [lindex $entry 0]
			set pos [lindex $entry 1]
			set mode [lindex $entry 2]
			set opts [lindex $entry 4]
			if {$n > -1} {
				load_file $n $fname $opts "set mode(\$no) \{$mode\}; setcursor \$w \{$pos\}"
				set nfiles [expr $n+1]
			} else {
				new_file $fname $opts "set mode(\$no) \{$mode\}; setcursor \$w \{$pos\}"
				set nfiles [expr $files+1]
			}
		}
	}
	project_make_menu
	if {$oldfiles >= $nfiles} {
		showinfo "Closing all old files ..."
		for {set i $nfiles} {$i < $oldfiles} {incr i} {
			if {$filename($i) != {}} {
				.nb raise $i
				file_close
			}
		}
	}
	if {$active != {}} {
		file_goto_named $active
	}
	showinfo "Opening new project <[file tail $file]> ... done."
}

proc project_add_file {} {
	global fno project filename mode

	if {$filename($fno) == {(Noname)}} {
		showinfo "Name the file first!"
		return
	}
	foreach e $project(files) {
		if {[lindex $e 0] == $filename($fno)} {
			showinfo "File <[file tail $filename($fno)]> is allready in the project!"
			return
		}
	}
	set project(saved) 0
	project_add_entry [list $filename($fno) [[twidgetbyno $fno] index insert] $mode($fno) 0 [get_file_options $fno]]
	project_make_menu
	project_save
	showinfo "Added file <[file tail $filename($fno)]> to the project."
}

proc project_del_file {} {
	global fno project filename

	for {set i 0} {$i < [llength $project(files)]} {incr i} {
		set name [lindex [lindex $project(files) $i] 0]
		if {[file tail $name] == [file tail $filename($fno)]} {
			set pfiles [lrange $project(files) 0 [expr $i-1]]
			append pfiles " [lrange $project(files) [expr $i+1] end]"
			set project(files) $pfiles
			set project(saved) 0
			project_make_menu
			project_save
			showinfo "Deleted file <[file tail $filename($fno)]> from the project."
			return
		}
	}
	showinfo "File <[file tail $filename($fno)]> is not in the project!"
}

proc project_goto_file {no} {
	global project files filename

	if {$no >= [llength $project(files)]} {
		showinfo "Project index out of project range!"
		return
	}
	file_goto_named [lindex [lindex $project(files) $no] 0]
}

proc project_sort_new {} {
	global project

	set pfiles $project(files)
	set project(files) {}
	foreach entry $pfiles {
		project_add_entry $entry
	}
	set project(saved) 0
	project_make_menu
	project_save
}

proc project_save {} {
	global	project

	if {$project(name) != {}} {
		project:save_file $project(name)
	}
}

proc project_saveas {} {

	file_dialog {project save} prj {project:save_file} {}
}

proc project:save_file {file} {
	global	project fno mode filename

	if {[string trim $file] != {}} {
		set project(name) [completename [file rootname $file]]
		set project(info) [file rootname [file tail $project(name)]]
		showinfo "Saving project file <[file tail $project(name)].prj> ..."
		set f [open $project(name).prj w]
		puts $f "geometry: [wm geometry .]"
		puts $f "sort: $project(sort)"
		if {$filename($fno) != {} && $filename($fno) != {(Noname)}} {
			puts $f "active: $filename($fno)"
		}
		foreach entry $project(files) {
			set line {}
			if {[set no [nobyname [lindex $entry 0]]] != -1} {
				lappend line [[twidgetbyno $no] index insert] $mode($no) 1
			} else {
				lappend line [lindex $entry 1] [lindex $entry 2] 0
			}
			puts $f "file: [lindex $entry 0] $line [get_file_options $fno]"
		}
		close $f
		set project(saved) 1
		showinfo "Saving project file <[file tail $project(name)].prj> ... done."
	} else {
		project_saveas
		return
	}
}

proc project_open {file} {

	file_dialog {file open} {prj} {project:open_file} $file
}

proc project:open_file {file} {
	global	filename mode

	if {$file != {}} {
		project_new $file
	}
	setfocus
}

proc project_close {} {
	global program filename fno project prjname
	
	set project(name)	{}
	set project(info]	{}
	set prjname		{}
	project_del_menu
	update idletasks

}
