
proc setcursor {w pos} {
	$w mark set insert $pos
	$w see insert
}

proc init_undo {no} {
	global	undo undovars

	set undo($no,0)		{}
	set undovars(count,$no)	0
	set undovars(begin,$no)	1.0
	set undovars(end,$no)	1.0
	set undovars(text,$no)	{}
	set undovars(del,$no)	-1
	set undovars(max,$no)	0
}

proc setundo {} {
	global	fno undo undovars

	set undo($fno,$undovars(count,$fno)) [list $undovars(del,$fno) $undovars(begin,$fno) $undovars(end,$fno) $undovars(text,$fno)]
}

proc newundo {del begin end text} {
	global	fno undo undovars

	setundo
	set undovars(max,$fno) [incr undovars(count,$fno)]
	set undovars(del,$fno) $del
	set undovars(begin,$fno) $begin
	set undovars(end,$fno) $end
	set undovars(text,$fno) $text
}

proc appendundo {del begin end text} {
	global	fno undo undovars

	if {$undovars(del,$fno) != $del} {
		newundo $del $begin $end $text
	} else {
		if {$undovars(del,$fno)} {
			if {$undovars(begin,$fno) == $begin} {
				append undovars(text,$fno) $text
				set undovars(end,$fno) $end
			} elseif {$undovars(begin,$fno) == $end} {
				set undovars(text,$fno) [append text $undovars(text,$fno)]
				set undovars(begin,$fno) $begin
			} else {
				newundo $del $begin $end $text
			}
		} else {
			if {$undovars(end,$fno) == $begin} {
				append undovars(text,$fno) $text
				set undovars(end,$fno) $end
			} else {
				newundo $del $begin $end $text
			}
		}
	}
}

proc doundo {} {
	global	fno undo undovars withundo

	if {$undovars(del,$fno) != -1} {
		set w [twidgetbyno $fno]
		if {$undovars(count,$fno) == $undovars(max,$fno)} {
			setundo
		}
		set withundo 0
		if {$undovars(del,$fno)} {
			setcursor $w $undovars(begin,$fno)
			puttext $w $undovars(text,$fno)
		} else {
			delregion $w $undovars(begin,$fno) $undovars(end,$fno)
		}
		setcursor $w $undovars(begin,$fno)
		set withundo 1
		incr undovars(count,$fno) -1
		if {$undovars(count,$fno) > 0} {
			set u $undo($fno,$undovars(count,$fno))
			set undovars(del,$fno) [lindex $u 0]
			set undovars(begin,$fno) [lindex $u 1]
			set undovars(end,$fno) [lindex $u 2]
			set undovars(text,$fno) [lindex $u 3]
		} else {
			set undovars(del,$fno) -1
			setmodified $fno 0
		}
	} else {
		showinfo "There is nothing to undo!"
	}
}

proc doredo {} {
	global	fno undo undovars withundo

	if {$undovars(count,$fno) < $undovars(max,$fno)} {
		set w [twidgetbyno $fno]
		incr undovars(count,$fno)
		set u $undo($fno,$undovars(count,$fno))
		set undovars(del,$fno) [lindex $u 0]
		set undovars(begin,$fno) [lindex $u 1]
		set undovars(end,$fno) [lindex $u 2]
		set undovars(text,$fno) [lindex $u 3]
		set withundo 0
		if {$undovars(del,$fno)} {
			delregion $w $undovars(begin,$fno) "$undovars(begin,$fno)+[string length $undovars(text,$fno)]c"
			setcursor $w $undovars(begin,$fno)
		} else {
			setcursor $w $undovars(begin,$fno)
			puttext $w $undovars(text,$fno)
		}
		set withundo 1
	} else {
		showinfo "There is nothing to redo!"
	}
}

proc deletetags {w} {

	$w tag delete [$w tag names]
}

proc tagcolors {w no} {
	global	options fileefi

	foreach tag $fileefi($no.tags) {
		foreach option [array names fileefi $no.tag.$tag.*] {
			set opt [extension $option]
			if {$opt != {del} && $opt != {info}} {
				if {$tag == {default}} {
					catch {$w configure -$opt $fileefi($option)}
				} else {
					catch {$w tag configure $tag -$opt $fileefi($option)}
				}
			}
		}
	}
	foreach type [array names options *.tags] {
		foreach tag $options($type) {
			foreach option [array names options *tag.$tag.*] {
				set opt [extension $option]
				if {$opt != {del} && $opt != {info}} {
					$w tag configure $tag -$opt $options($option)
				}
			}
		}
	}
}

proc newremark w {
	global	fno mode

	if {$mode($fno) == {C} || $mode($fno) == {JAVA} || $mode($fno) == {DSP}} {
		$w tag remove rem1 1.0 end
		$w mark set last 1.0
		while {[set ind [$w search {/*} last end]] != {}} {
			set inde [$w search {*/} $ind+2c end]
			if {$inde == {}} {
				set inde end
			}
			$w tag add rem1 $ind $inde+2c
			$w mark set last $inde+2c
			if {$inde == {end}} break;
		}
	}
	if {$mode($fno) == {DSP}} {
		$w tag remove rem2 1.0 end
		$w mark set last 1.0
		while {[set ind [$w search "\{" last end]] != {}} {
			set inde [$w search "\}" $ind+1c end]
			if {$inde == {}} {
				set inde end
			}
			$w tag add rem2 $ind $inde+1c
			$w mark set last $inde+1c
			if {$inde == {end}} break;
		}
	}
}

proc delregion {w pos1 pos2} {
	global	fno mode withundo

	set rem 0
	setmodified $fno 1
	set ch [$w get $pos1 $pos2]
	set p1 [$w index $pos1]
	set p2 [$w index $pos2]
	if {$withundo} {
		appendundo 1 $p1 $p2 $ch
	}
	if {$mode($fno) == {C} || $mode($fno) == {JAVA} || $mode($fno) == {DSP}} {
		set c1 [$w get "$p1-1c" $p1]
		set c2 [$w get $p2 "$p2+1c"]
		if {[$w search {/*} $p1-1c $p2+1c] != {} || [$w search {*/} $p1-1c $p2+1c] != {}} {
			set rem 1
		} elseif {($c1 == {*} && $c2 == {/}) || ($c1 == {/} && $c2 == {*})} {
			set rem 1
		} elseif {$mode($fno) == {DSP}} {
			if {[$w search -regexp {[{}]} $p1 $p2] != {}} {
				set rem 1
			}
		}
		$w delete $p1 $p2
		if {$rem} {
			newremark $w
		}
	} else {
		$w delete $pos1 $pos2
	}
	colorize $w $pos1 [$w index "$pos1+1l linestart"]
	$w see $pos1
	return $ch
}

proc puttext {w text} {
	global	fno mode withundo

	set rem 0
	set i1 [$w index insert]
	$w insert insert $text
	set i2 [$w index insert]
	if {$withundo} {
		appendundo 0 $i1 $i2 $text
	}
	if {$mode($fno) == {C} || $mode($fno) == {JAVA} || $mode($fno) == {DSP}} {
		set c1 [$w get "$i1-1c" $i1]
		set c2 [$w get $i2 "$i2+1c"]
		if {[$w search {/*} $i1-1c $i2+1c] != {} || [$w search {*/} $i1-1c $i2+1c] != {}} {
			set rem 1
		} elseif {($c1 == {*} && $c2 == {/}) || ($c1 == {/} && $c2 == {*})} {
			set rem 1
		} elseif {$text == {*} && ($c1 == {/} || $c2 == {/})} {
			set rem 1
		} elseif {$text == {/} && ($c1 == {*} || $c2 == {*})} {
			set rem 1
		} elseif {$mode($fno) == {DSP}} {
			if {[$w search -regexp {[{}]} $i1 $i2] != {}} {
				set rem 1
			}
		}
	}
	if {$rem} {
		newremark $w
	}
	if {$i1 != $i2 || $rem} {
		setmodified $fno 1
		colorize $w $i1 [$w index "$i2+1l linestart"]
	}
	$w see insert
}

proc blinkbrace {w pos brace} {
	global	bracefind fno

	switch $brace {
		\{	{ set findbs {[{}]}; set bs "\}"; set dir {-forwards}; set bn brace }
		\}	{ set findbs {[{}]}; set bs "\{"; set dir {-backwards}; set bn brace }
		\[	{ set findbs {[][]}; set bs "\]"; set dir {-forwards}; set bn bracket }
		\]	{ set findbs {[][]}; set bs "\["; set dir {-backwards}; set bn bracket }
		(	{ set findbs {[()]}; set bs ")"; set dir {-forwards}; set bn parencesis }
		)	{ set findbs {[()]}; set bs "("; set dir {-backwards}; set bn parencesis }
	}
	set p [set i [$w index $pos-1c]]
	set d 1
# The next 3 lines make sure that eddi won't give an error when you type
# / AltGr+)
	if {![info exists dir]} {
		return
	}
	while {$d > 0} {
		if {$dir == "-backwards"} {
			set p [$w search $dir -regexp $findbs $p 1.0]
		} else {
			set p [$w search $dir -regexp $findbs $p+1c end]
		}
		if {$p == {}} {
			set d -1
		} else {
			if {"[$w get $p $p+1c]" == "$bs"} {
				incr d -1
			} else {
				incr d 1
			}
		}
	}
	if {$d == 0} {
		if {$dir == "-backwards"} {
			$w tag add sel $p $pos
		} else {
			$w tag add sel $pos-1c $p+1c
		}
		$w see $p
		update idletasks
		after 300
		$w tag remove sel 1.0 end
		$w see insert
	} else {
		showinfo "No corresponding <$bs> found!"
	}
}

proc showifbrace {w} {

	if {[$w tag ranges sel] != {}} {
		set i1 [$w index sel.first]
		set i2 [$w index sel.last]
		if {[regexp {[][{}()]} [set c [$w get sel.first sel.last]]] && [string length $c] == 1} {
			blinkbrace $w sel.last $c
		}
		$w tag add sel $i1 $i2
	}
}

proc insblinkbrace {w brace} {
	global	fileopts fno macro

	puttext $w $brace
	if {$fileopts($fno.bracefind) && !$macro(play)} {
		blinkbrace $w insert $brace
	}
}

proc search_backup {} {
	global	search

	set search(cdir) $search(dir); set search(dir) {-forwards}
	set search(ccase) $search(case); set search(case) {-nocase}
	set search(ctype) $search(type); set search(type) {-exact}
	set search(csearch) $search(search); set search(search) {}
	set search(creplace) $search(replace); set search(replace) {}
}

proc search_restore {} {
	global	search

	set search(dir) $search(cdir)
	set search(case) $search(ccase)
	set search(type) $search(ctype)
	set search(search) $search(csearch)
	set search(replace) $search(creplace)
}

proc findtext {w} {
	global	search

	if {$search(search) != {}} {
		$w tag remove sel 1.0 end
		set count 0
		if {$search(dir) == {-forwards}} {
			set start {insert+1c}
			set ende end
		} else {
			set start {insert}
			set ende 1.0
		}
		set ind [eval $w search -count count $search(dir) $search(case) $search(type) -- "[list $search(search)]" $start $ende]
		if {$ind != {}} {
			$w tag add sel $ind "$ind+$count c"
			setcursor $w $ind
			focus $w
		}
		return "$ind $count"
	}
}

proc replacetext {w} {
	global	search

	if {[$w tag ranges sel] != {}} {
		set p [$w index sel.first]
		$w mark set insert $p
		delregion $w [$w index sel.first] [$w index sel.last]
		puttext $w $search(replace)
		return 1
	}
	return 0
}

proc replacefindtext {w} {
	global	search

	if {$search(search) != {}} {
		replacetext $w
		return [findtext $w]
	}
}

proc replacefindalltext {w} {
	global	search

	if {$search(search) != {}} {
		while {[lindex [replacefindtext $w] 1] > 0} {}
	}
}

proc gotoline {line} {
	global	fno tkPriv

	set w [twidgetbyno $fno]
	if {$line == {}} {
		if {[input_panel .egl gotoline {Goto line} 10 {} {}]} {
			if {[set gtl [string trim $tkPriv(input)]] != {}} {
				command [list goto line $gtl]
			}
		}
		return
	}
	if {[regexp {[1-9][0-9]*} $line]} {
		$w mark set insert "$line.0"
		showposition $w
		$w see insert
	} else {
		showinfo "Bad line number <$line>!"
	}
}

proc copyselection {w} {
	global	cutbuffer

	if {[$w tag ranges sel] != {}} {
		set cutbuffer [$w get sel.first sel.last]
	}
}

proc cutselection {w} {
	global	cutbuffer

	if {[$w tag ranges sel] != {}} {
		catch {set cutbuffer [$w get sel.first sel.last]}
		catch {delregion $w [$w index sel.first] [$w index sel.last]}
	}
}

proc selection_indent {w text} {

	if {[selection_exists $w]} {
		set range [selection_range $w]
		scan [lindex $range 0] "%d.%d" y1 x1
		if {$x1 > 0} {
			incr y1
		}
		scan [lindex $range 1] "%d.%d" y2 x2
		if {$x2 == 0} {
			incr y2 -1
		}
		for {set l $y1} {$l <= $y2} {incr l} {
			$w insert $l.0 "$text"
		}
		if {$x1 == 0} {
			$w tag add sel $y1.0 $y1.1
		}
	}
}

proc selection_reindent {w text} {

	if {[selection_exists $w]} {
		set range [selection_range $w]
		scan [lindex $range 0] "%d.%d" y1 x1
		if {$x1 > 0} {
			incr y1
		}
		scan [lindex $range 1] "%d.%d" y2 x2
		if {$x2 == 0} {
			incr y2 -1
		}
		for {set l $y1} {$l <= $y2} {incr l} {
			if {[$w get $l.0 "$l.0+[string length $text]c"] == "$text"} {
				$w delete $l.0 "$l.0+[string length $text]c"
			}
		}
	}
}

proc selection_remark {w} {
	global	mode fno efi

	if {$mode($fno) == {}} {
		showinfo "No texttype specified."
		return
	}	
	if {[selection_exists $w]} {
		set old [$w index insert]
		set range [selection_range $w]
		scan [lindex $range 0] "%d.%d" y1 x1
		if {$x1 > 0} {
			incr y1
		}
		scan [lindex $range 1] "%d.%d" y2 x2
		if {$x2 == 0} {
			incr y2 -1
		}
		set remark $efi($mode($fno))
		if {"[lindex $remark 0]" == "[lindex $remark 1]" && "[lindex $remark 1]" == "[lindex $remark 2]"} {
			set range 0
		} else {
			showinfo "Not available vor the $mode($fno)-Type!"
			return
		}
		for {set l $y1} {$l <= $y2} {incr l} {
			set rem 0
			foreach t $range {
				set text [lindex $remark $t]
				if {[set p [$w search -exact -count num -- $text $l.0 "$l.0 lineend"]] != {}} {
					if {[string trim [$w get $l.0 "$p+$num c"]] == "$text"} {
						set rem 1
						break
					}
				}
			}
			if {!$rem} {
				$w mark set insert $l.0
				if {$l == $y1} {
					puttext $w [lindex $remark 0]
				} elseif {$l == $y2} {
					puttext $w [lindex $remark 2]
				} else {
					puttext $w [lindex $remark 1]
				}
			}
		}
		if {$x1 == 0} {
			$w tag add sel $y1.0 "$y1.0+[string length [lindex $remark 0]]c"
		}
		$w mark set insert $old
	}
}

proc selection_delremark {w} {
	global	mode fno efi

	if {$mode($fno) == {}} {
		showinfo "No texttype specified."
		return
	}	
	if {[selection_exists $w]} {
		set range [selection_range $w]
		scan [lindex $range 0] "%d.%d" y1 x1
		if {$x1 > 0} {
			incr y1
		}
		scan [lindex $range 1] "%d.%d" y2 x2
		if {$x2 == 0} {
			incr y2 -1
		}
		set remark $efi($mode($fno))
		if {"[lindex $remark 0]" == "[lindex $remark 1]" && "[lindex $remark 1]" == "[lindex $remark 2]"} {
			set range 0
		} else {
			showinfo "Not available vor the $mode($fno)-Mode!"
			return
		}
		for {set l $y1} {$l <= $y2} {incr l} {
			foreach t $range {
				set text [lindex $remark $t]
				if {[set p [$w search -exact -count num -- $text $l.0 "$l.0 lineend"]] != {}} {
					if {[string trim [$w get $l.0 "$p+$num c"]] == "$text"} {
						delregion $w $p "$p+$num c"
						break
					}
				}
			}
		}
	}
}

proc inserttab {w} {
	puttext $w \t
}

proc insertnewline {w} {
	global	fileopts fno

	set n {}
	if {$fileopts($fno.autoindent)} {
		$w mark set p1 {insert linestart}
		set c [$w get p1 {p1+1c}]
		while {$c == { } || $c == "\t"} {
			$w mark set p1 {p1+1c}
			set c [$w get p1 {p1+1c}]
		}
		set n [$w get {insert linestart} p1]
		$w mark unset p1
	}
	puttext $w "\n$n"
	if {$fileopts($fno.autoindent)} {
		set c [$w get insert "insert+1c"]
		while {$c == " " || $c == "\t"} {
			$w delete insert "insert+1c"
			set c [$w get insert "insert+1c"]
		}
	}
}

proc grephelp {w} {
	global	helptext

	if {![regexp {[a-zA-Z0-9_]} [$w get "insert-1c" insert]]} {
		set helptext [string trim [$w get "insert+1c wordstart" "insert+1c wordend"]]
	} else {
		set helptext [string trim [$w get "insert-1c wordstart" "insert-1c wordend"]]
	}
	if {$helptext != ""} {
		showhelp $helptext
	}
}

proc deletenextword {w pos} {

	if {[regexp "\[ \t\n\]" [$w get $pos $pos+1c]]} {
		set p1 [$w search -regexp "\[^ \t\n\]" $pos end]
		if {[$w compare $p1 > $pos]} {
			delregion $w $pos $p1
		}
	} else {
		delregion $w $pos "$pos wordend"
	}
}

proc indexprevword {w pos} {

	$w mark set p1 $pos
	while {![regexp {[a-zA-Z0-9_]} [$w get "p1-1c" p1]] && [$w index p1] != 1.0} {
		$w mark set p1 "p1-1c"
	}
	while {[regexp {[a-zA-Z0-9_]} [$w get "p1-1c" p1]] && [$w index p1] != 1.0} {
		$w mark set p1 "p1-1c"
	}
	set p [$w index p1]
	$w mark unset p1
	return $p
}

proc indexnextword {w pos} {

	$w mark set p1 $pos
	while {[regexp {[a-zA-Z0-9_]} [$w get p1 "p1+1c"]] && [$w index p1] != [$w index end]} {
		$w mark set p1 "p1+1c"
	}
	while {![regexp {[a-zA-Z0-9_]} [$w get p1 "p1+1c"]] && [$w index p1] != [$w index end]} {
		$w mark set p1 "p1+1c"
	}
	set p [$w index p1]
	$w mark unset p1
	return $p
}

proc filter_text {prog} {
	global	fno tkPriv

	set w [twidgetbyno $fno]
	if {[selection_exists $w]} {
		set what selection
	} else {
		set what file
	}
	if {$prog == {}} {
		if {[input_panel .eft "filter $what" Command 30 {} {}]} {
			if {[set prog [string trim $tkPriv(input)]] != {}} {
				command [list insert filter $prog]
			}
		}
		return
	}
	showinfo "Filter $what to <$prog> ..."
	if {[selection_exists $w] || $what == {file}} {
		if {$what == {file}} {
			set t [delregion $w 1.0 end-1c]
			$w delete 1.0 end
		} else {
			$w mark set insert [lindex [selection_range $w] 0]
			set p1 [$w index insert]
			$w see insert
			set t [selection_del $w]
		}
		if {![catch {set f [eval exec $prog <<\$t]}]} {
			append f "\n"
			puttext $w $f
			if {!($what == {file})} {
				$w tag add sel $p1 insert
			}
			showinfo "Filter $what to <$prog> ... done."
		} else {
			puttext $w $t
			showinfo "Filter $what to <$prog> ... failed."
		}
	} else {
		showinfo "No selecion."
	}
}

proc exec_prog {prog} {
	global	fno tkPriv

	set w [twidgetbyno $fno]
	if {$prog == {}} {
		if {[input_panel .eep "execute command" Command 40 {} {}]} {
			if {[set prog [string trim $tkPriv(input)]] != {}} {
				command [list insert exec $prog]
			}
		}
		return
	}
	showinfo "Execute <$prog> ..."
	if {![catch {set f [eval exec $prog]}]} {
		puttext $w $f
		showinfo "Execute <$prog> ... done."
	} else {
		showinfo "Execute <$prog> ... failed."
	}
}

proc insert_privat {what} {
	global	fno filename

	set t {}
	switch -exact $what {
	name		{ set t [file tail $filename($fno)] }
	path		{ set t [file dirname $filename($fno)] }
	file		{ set t $filename($fno) }
	ext		{ set t [extension $filename($fno)]}
	root		{ set t [file tail [file rootname $filename($fno)]] }
	}
	puttext [twidgetbyno $fno] $t
}

proc bookmark_new {w name} {
	return [label [file rootname $w].bm$name -text $name -bg yellow -bd 0]
}

proc bookmark_list {w} {

	set l {}
	foreach mark [$w mark names] {
		if {[regexp bm.* $mark]} {
			lappend l [string range $mark 2 end]
		}
	}
	return $l
}

proc bookmark_set {w name pos} {
	global fileopts modified tkPriv

	set no [nobytwidget $w]
	if {[string trim $name] == {}} {
		if {[input_panel .eep "bookmark set" Bookmark 15 {} [bookmark_list $w]]} {
			if {[set name [string trim $tkPriv(input)]] != {}} {
				command [list insert bookmark $name]
			}
		}
		return
	}
	if {![regexp {[A-Za-z0-9_]*} $name]} {
		showinfo "Bookmark name <$name> not allowed!"
		return
	}
	setmodified $no 1
	catch {destroy [file rootname $w].bm$name}
	if {$fileopts($no.showbms)} {
		$w window create $pos -window [bookmark_new $w $name]
		catch {$w mark set bm$name [file rootname $w].bm$name}
	} else {
		$w mark set bm$name $pos
	}
	showinfo "Set bookmark <$name>."
}

proc bookmark_defined {w name} {
	global	fileopts

	set no [nobytwidget $w]
	if {[lsearch [$w mark names] bm$name] != -1} {
		if {[lsearch [$w window names] [file rootname $w].bm$name] != -1} {
			return 1
		} else {
			if {$fileopts($no.showbms)} {
				return 0
			}
			return 1
		}
	} else {
		return 0
	}
}

proc bookmark_goto {w name} {
	global	fileopts tkPriv

	set no [nobytwidget $w]
	if {[string trim $name] == {}} {
		if {[input_panel .eep "bookmark goto" Bookmark 15 {} [bookmark_list $w]]} {
			if {[set name [string trim $tkPriv(input)]] != {}} {
				command [list goto bookmark $name]
			}
		}
		return
	}
	if {![regexp {[A-Za-z0-9_]*} $name]} {
		showinfo "Bookmark name <$name> not allowed!"
		return
	}
	if {[bookmark_defined $w $name]} {
		if {$fileopts($no.showbms)} {
			setcursor $w bm$name+1c
		} else {
			setcursor $w bm$name
		}
		showinfo "Goto bookmark <$name>."
	} else {
		showinfo "No bookmark <$name> defined."
	}
}

proc insert_special_char {w num} {

	if {[string trim $num] == {}} {
		special_char_panel .scp
		return
	}
	if {![catch {set ch [format "%c" $num]}]} {
		puttext $w $ch
	} else {
		showinfo "Bad number <$num> for special character!"
	}
}

proc selection_tag {w do tags} {
	global	options tagstate fno fileopts fileefi

	if {[string trim $tags] == {}} {
		selection_tag_panel .est
		return
	}
	if {![selection_exists $w]} {
		showinfo "No selection to tag!"
		return
	}
	set sel(w) [selection_w $w]
	set sel(first) [lindex [selection_range $w] 0]
	set sel(last) [lindex [selection_range $w] 1]
	if {$tags == {all}} {
		set tags "$options(user.tags) $fileefi($fno.tags)"
	}
	foreach tag $tags {
		if {[lsearch $options(user.tags) $tag] == -1} {
			if {[lsearch $fileefi($fno.tags) $tag] == -1} {
				continue
			}
			set ttype file
		} else {
			set ttype user
		}
		set d $do
		if {$do != {tag} && $do != {untag}} {
			if {[lsearch -exact [$w tag names $sel(first)] $tag] == -1} {
				set d tag
			} else {
				set d untag
			}
		}
		set delete {}
		if {$ttype == {user}} {
			foreach e [array names options tag.$tag.del] {
				set delete $options($e)
			}
		} else {
			foreach e [array names fileefi $fno.tag.$tag.del] {
				set delete $fileefi($e)
			}
		}
		foreach del $delete {
			$sel(w) tag remove $del $sel(first) $sel(last)
		}
		if {$d == {tag}} {
			$w tag add $tag $sel(first) $sel(last)
		} else {
			$w tag remove $tag $sel(first) $sel(last)
		}
	}
}

proc selection_send {w name} {
	global	tkPriv remote

	if {![selection_exists $w]} {
		showinfo {There is no selection to send!}
		return
	}
	if {[string trim $name] == {}} {
		if {[input_panel .ess "selection send" Interpreter 15 $remote(name) [winfo interps]]} {
			if {[set name [string trim $tkPriv(input)]] != {}} {
				set remote(name) $name
				command [list selection send $name]
			}
		}
		return
	}
	set remote(w) $w
	uplevel #0 {catch {send $remote(name) [selection_get $remote(w)]}}
}

proc selection_w {w} {
	global	wssel

	if {[$w tag ranges sel] != {}} {
		return $w
	} else {
		if {[wssel_visible]} {
			return $wssel(w)
		}
	}
	return
}

proc selection_range {w} {
	global	wssel

	if {[$w tag ranges sel] != {}} {
		return "[$w index sel.first] [$w index sel.last]"
	} else {
		if {[wssel_visible]} {
			return "[$wssel(w) index wssel.first] [$wssel(w) index wssel.last]"
		}
	}
	return
}

proc selection_exists {w} {

	if {[$w tag ranges sel] != {}} {
		return 1
	} else {
		return [wssel_visible]
	}
}

proc selection_get {w} {

	if {![selection_exists $w]} return
	if {[$w tag ranges sel] != {}} {
		return [$w get sel.first sel.last]
	} else {
		return [wssel_get]
	}
}

proc selection_del {w} {

	if {![selection_exists $w]} return
	if {[$w tag ranges sel] != {}} {
		return [delregion $w [$w index sel.first] [$w index sel.last]]
	} else {
		return [wssel_del]
	}
}

proc wssel_visible {} {
	global	wssel

	if {$wssel(has) == {} || [lsearch [.nb pages] $wssel(has)] == -1} {
		return 0
	} elseif {[$wssel(w) tag ranges wssel] == {}} {
		return 0
	}
	return 1
}

proc wssel_off {} {
	global	wssel

	if {![wssel_visible]} return
	$wssel(w) tag remove wssel 1.0 end
	set wssel(w) {}
	set wssel(has) {}
}

proc wssel_toggle {} {
	global	wssel

	if {[wssel_visible]} {
		$wssel(w) tag remove wssel 1.0 end
	} else {
		if {$wssel(has) != {} && [lsearch [.nb pages] $wssel(has)] != -1} {
			$wssel(w) tag add wssel wsbegin wsend
		}
	}
}

proc wssel_begin {w pos} {
	global	wssel

	if {$wssel(has) == {} || $wssel(w) != $w} {
		wssel_off
		set wssel(w) $w
		set wssel(has) [nobytwidget $w]
		$w mark set wsbegin $pos
		$w mark gravity wsbegin right
		catch {$w mark unset wsend}
	} else {
		$w tag remove wssel 1.0 end
		$w mark set wsbegin $pos
		$w mark gravity wsbegin right
		catch {$w tag add wssel wsbegin wsend}
	}
}

proc wssel_end {w pos} {
	global	wssel

	if {$wssel(has) == {} || $wssel(w) != $w} {
		wssel_off
		set wssel(w) $w
		set wssel(has) [nobytwidget $w]
		$w mark set wsend $pos
		$w mark gravity wsend left
		catch {$w mark unset wsbegin}
	} else {
		$w tag remove wssel 1.0 end
		$w mark set wsend $pos
		$w mark gravity wsend left
		catch {$w tag add wssel wsbegin wsend}
	}
}

proc wssel_get {} {
	global	wssel

	if {![wssel_visible]} return
	return [$wssel(w) get wssel.first wssel.last]
}

proc wssel_del {} {
	global	wssel fno

	if {![wssel_visible]} return
	set ono $fno
	set fno $wssel(has)
	set t [delregion $wssel(w) [$wssel(w) index wssel.first] [$wssel(w) index wssel.last]]
	set wssel(has) {}
	set wssel(w) {}
	set fno $ono
	return $t
}

proc wssel_copy {w} {

	if {![wssel_visible]} return
	set p [$w index insert]
	puttext $w [wssel_get]
	wssel_begin $w $p
	wssel_end $w insert
}

proc wssel_move {w} {

	if {![wssel_visible]} return
	set t [wssel_del]
	set p [$w index insert]
	puttext $w $t
	wssel_begin $w $p
	wssel_end $w insert
}
